@@ -1,65 +1,98 @@
-import copy
+import os
 import csv
 import json
 import gzip
-import pickle
 import logging
 import itertools
-from collections import defaultdict
+from pathlib import Path
+from collections import defaultdict, Counter
+from textwrap import dedent
+from typing import List, Mapping, Optional, Set, Tuple, Union
 from adeft.disambiguate import load_disambiguator
 from adeft.modeling.classify import load_model_info
 from adeft import available_shortforms as available_adeft_models
-from .term import Term
+from .term import Term, get_identifiers_curie, get_identifiers_url
 from .process import normalize, replace_dashes, replace_greek_uni, \
     replace_greek_latin, replace_greek_spelled_out, depluralize, \
     replace_roman_arabic
-from .scorer import generate_match, score, score_namespace
+from .scorer import Match, generate_match, score, score_namespace
 from .resources import get_gilda_models, get_grounding_terms
 
+__all__ = [
+    "Grounder",
+    "GrounderInput",
+    "ScoredMatch",
+    "load_terms_file",
+    "filter_for_organism",
+    "load_adeft_models",
+    "load_gilda_models",
+]
 
 logger = logging.getLogger(__name__)
 
 
+GrounderInput = Union[str, Path, List[Term], Mapping[str, List[Term]]]
+
+
 class Grounder(object):
     """Class to look up and ground query texts in a terms file.
 
     Parameters
     ----------
-    terms : str or dict or None
+    terms :
         Specifies the grounding terms that should be loaded in the Grounder.
-        If None, the default grounding terms are loaded from the versioned
-        resource folder. If str, it is interpreted as a path to a grounding
-        terms TSV file which is then loaded. If dict, it is assumed to be
-        a grounding terms dict with normalized entity strings as keys
-        and Term objects as values. Default: None
+
+        - If ``None``, the default grounding terms are loaded from the
+          versioned resource folder.
+        - If :class:`str` or :class:`pathlib.Path`, it is interpreted
+          as a path to a grounding terms gzipped TSV file which is then
+          loaded.
+        - If :class:`list`, it is assumed to be a flat list of
+          :class:`gilda.term.Term` instances.
+        - If :class:`dict`, it is assumed to be a grounding terms dict with
+          normalized entity strings as keys and :class:`gilda.term.Term`
+          instances as values.
     """
-    def __init__(self, terms=None):
+
+    entries: Mapping[str, List[Term]]
+
+    def __init__(self, terms: Optional[GrounderInput] = None):
         if terms is None:
             terms = get_grounding_terms()
 
-        if isinstance(terms, str):
-            self.entries = load_terms_file(terms)
+        if isinstance(terms, (str, Path)):
+            extension = os.path.splitext(terms)[1]
+            if extension == '.db':
+                from .resources.sqlite_adapter import SqliteEntries
+                self.entries = SqliteEntries(terms)
+            else:
+                self.entries = load_terms_file(terms)
+        elif isinstance(terms, list):
+            self.entries = defaultdict(list)
+            for term in terms:
+                self.entries[term.norm_text].append(term)
+            self.entries = dict(self.entries)
         elif isinstance(terms, dict):
             self.entries = terms
         else:
-            raise TypeError('terms is neither a path nor a normalized'
-                            ' entry name to term dictionary')
+            raise TypeError('terms is neither a path nor a list of terms,'
+                            'nor a normalized entry name to term dictionary')
 
-        self.adeft_disambiguators = load_adeft_models()
-        self.gilda_disambiguators = load_gilda_models()
+        self.adeft_disambiguators = find_adeft_models()
+        self.gilda_disambiguators = None
 
-    def lookup(self, raw_str):
+    def lookup(self, raw_str: str) -> List[Term]:
         """Return matching Terms for a given raw string.
 
         Parameters
         ----------
-        raw_str : str
+        raw_str :
             A string to be looked up in the set of Terms that the Grounder
             contains.
 
         Returns
         -------
-        list of Term
+        :
             A list of Terms that are potential matches for the given string.
         """
         lookups = self._generate_lookups(raw_str)
@@ -68,7 +101,7 @@ class Grounder(object):
             entries += self.entries.get(lookup, [])
         return entries
 
-    def _generate_lookups(self, raw_str):
+    def _generate_lookups(self, raw_str: str) -> Set[str]:
         # TODO: we should propagate flags about depluralization and possible
         #  other modifications made here and take them into account when
         #  scoring
@@ -97,7 +130,8 @@ class Grounder(object):
                      ', '.join(lookups))
         return lookups
 
-    def ground(self, raw_str, context=None, organisms=None):
+    def ground(self, raw_str, context=None, organisms=None,
+               namespaces=None):
         """Return scored groundings for a given raw string.
 
         Parameters
@@ -109,6 +143,17 @@ class Grounder(object):
             Any additional text that serves as context for disambiguating the
             given entity text, used if a model exists for disambiguating the
             given text.
+        organisms : Optional[List[str]]
+            An optional list of organism identifiers defining a priority
+            ranking among organisms, if genes/proteins from multiple
+            organisms match the input. If not provided, the default
+            ['9606'] i.e., human is used.
+        namespaces : Optional[List[str]]
+            A list of namespaces to restrict matches to. This will apply to
+            both the primary namespace of a matched term, to any subsumed
+            matches, and to the source namespaces of terms if they were
+            created using cross-reference mappings. By default, no
+            restriction is applied.
 
         Returns
         -------
@@ -151,9 +196,21 @@ class Grounder(object):
         rank_fun = lambda x: (x.score, score_namespace(x.term))
         unique_scores = sorted(unique_scores, key=rank_fun, reverse=True)
 
+        # If we have a namespace constraint, we filter to the given
+        # namespaces.
+        if namespaces:
+            unique_scores = [
+                scored_match for scored_match in unique_scores
+                if scored_match.get_namespaces() & set(namespaces)
+            ]
+
         return unique_scores
 
     def disambiguate(self, raw_str, scored_matches, context):
+        # This is only called if context was passed in so we do lazy
+        # loading here
+        if self.gilda_disambiguators is None:
+            self.gilda_disambiguators = load_gilda_models()
         # If we don't have a disambiguator for this string, we return with
         # the original scores intact. Otherwise, we attempt to disambiguate.
         if raw_str in self.adeft_disambiguators:
@@ -176,6 +233,8 @@ class Grounder(object):
     def disambiguate_adeft(self, raw_str, scored_matches, context):
         # We find the disambiguator for the given string and pass in
         # context
+        if self.adeft_disambiguators[raw_str] is None:
+            self.adeft_disambiguators[raw_str] = load_disambiguator(raw_str)
         res = self.adeft_disambiguators[raw_str].disambiguate([context])
         # The actual grounding dict is at this index in the result
         grounding_dict = res[0][2]
@@ -284,6 +343,99 @@ class Grounder(object):
                     names.add(entry.text)
         return sorted(names)
 
+    def get_ambiguities(self,
+                        skip_names: bool = True,
+                        skip_curated: bool = True,
+                        skip_name_matches: bool = True,
+                        skip_species_ambigs: bool = True) -> List[List[Term]]:
+        """Return a list of ambiguous term groups in the grounder.
+
+        Parameters
+        ----------
+        skip_names :
+            If True, groups of terms where one has the "name" status are
+            skipped. This makes sense usually since these are prioritized over
+            synonyms anyway.
+        skip_curated :
+            If True, groups of terms where one has the "curated" status
+            are skipped. This makes sense usually since these are prioritized
+            over synonyms anyway.
+        skip_name_matches :
+            If True, groups of terms that all share the same standard name
+            are skipped. This is effective at eliminating spurious ambiguities
+            due to unresolved cross-references between equivalent terms
+            in different namespaces.
+        skip_species_ambigs :
+            If True, groups of terms that are all genes or proteins, and are
+            all from different species (one term from each species) are skipped.
+            This is effective at eliminating ambiguities between orthologous
+            genes in different species that are usually resolved using the
+            organism priority list.
+        """
+        ambig_entries = defaultdict(list)
+        for terms in self.entries.values():
+            for term in terms:
+                # We consider it an ambiguity if the same text entry appears
+                # multiple times
+                key = term.text
+                ambig_entries[key].append(term)
+
+        # It's only an ambiguity if there are two entries at least
+        ambig_entries = {k: v for k, v in ambig_entries.items()
+                         if len(v) >= 2}
+
+        ambigs = []
+        for text, entries in ambig_entries.items():
+            dbs = {e.db for e in entries}
+            db_ids = {(e.db, e.id) for e in entries}
+            statuses = {e.status for e in entries}
+            sources = {e.source for e in entries}
+            names = {e.entry_name for e in entries}
+            # If the entries all point to the same ID, we skip it
+            if len(db_ids) <= 1:
+                continue
+            # If there is a name in statuses, we skip it because it's
+            # prioritized
+            if skip_names and 'name' in statuses:
+                continue
+            # We skip curated terms because they are prioritized anyway
+            if skip_curated and 'curated' in statuses:
+                continue
+            # If there is an adeft model already, we skip it
+            if 'adeft' in sources:
+                continue
+            if skip_name_matches:
+                if len({e.entry_name.lower() for e in entries}) == 1:
+                    continue
+            if skip_species_ambigs:
+                if dbs <= {'HGNC', 'UP'} and \
+                        len({e.organism for e in entries}) == len(entries):
+                    continue
+            # Everything else is an ambiguity
+            ambigs.append(entries)
+        return ambigs
+
+    def _iter_terms(self):
+        for terms in self.entries.values():
+            yield from terms
+
+    def summary_str(self) -> str:
+        """Summarize the contents of the grounder."""
+        namespaces = {ns for term in self._iter_terms() for ns in term.get_namespaces()}
+        status_counter = dict(Counter(term.status for term in self._iter_terms()))
+        return dedent(f"""\
+        Lookups: {len(self.entries):,}
+        Terms: {sum(len(terms) for terms in self.entries.values()):,}
+        Term Namespaces: {namespaces}
+        Term Statuses: {status_counter}
+        Adeft Disambiguators: {len(self.adeft_disambiguators):,}
+        Gilda Disambiguators: {len(self.gilda_disambiguators):,}
+        """)
+
+    def print_summary(self, **kwargs) -> None:
+        """Print the summary of this grounder."""
+        print(self.summary_str(), **kwargs)
+
 
 class ScoredMatch(object):
     """Class representing a scored match to a grounding term.
@@ -305,7 +457,7 @@ class ScoredMatch(object):
         associated with a match can provide additional metadata in
         downstream applications.
     """
-    def __init__(self, term, score, match, disambiguation=None,
+    def __init__(self, term: Term, score, match: Match, disambiguation=None,
                  subsumed_terms=None):
         self.term = term
         self.url = term.get_idenfiers_url()
@@ -342,19 +494,59 @@ class ScoredMatch(object):
                      % (self.term.entry_name, value))
         self.score = self.score * value
 
+    def get_namespaces(self) -> Set[str]:
+        """Return all namespaces for this match including from mapped and
+        subsumed terms.
 
-def load_terms_file(terms_file):
+        Returns
+        -------
+        :
+            A set of strings representing namespaces for terms involved in
+            this match, including the namespace for the primary term as well
+            as any subsumed terms, and groundings that come from having
+            mapped an original source grounding during grounding resource
+            construction.
+        """
+        return {ns for ns, _ in self.get_groundings()}
+
+    def get_groundings(self) -> Set[Tuple[str, str]]:
+        """Return all groundings for this match including from mapped and
+        subsumed terms.
+
+        Returns
+        -------
+        :
+            A set of tuples representing groundings for this match including
+            the grounding for the primary term as well as any subsumed
+            terms, and groundings that come from having mapped an original
+            source grounding during grounding resource construction.
+        """
+        term_groundings = self.term.get_groundings()
+        if self.subsumed_terms:
+            for sub_term in self.subsumed_terms:
+                term_groundings |= sub_term.get_groundings()
+        return term_groundings
+
+    def get_grounding_dict(self) -> Mapping[str, str]:
+        """Get the groundings as CURIEs and URLs."""
+        return {
+            get_identifiers_curie(db, db_id): get_identifiers_url(db, db_id)
+            for db, db_id in self.get_groundings()
+        }
+
+
+def load_terms_file(terms_file: Union[str, Path]) -> Mapping[str, List[Term]]:
     """Load a TSV file containing terms into a lookup dictionary.
 
     Parameters
     ----------
-    terms_file : str
+    terms_file :
         Path to a TSV terms file with columns corresponding to the serialized
         elements of a Term.
 
     Returns
     -------
-    dict
+    :
         A lookup dictionary whose keys are normalized entity texts, and values
         are lists of Terms with that normalized entity text.
     """
@@ -393,16 +585,21 @@ def filter_for_organism(terms, organisms):
     return all_terms
 
 
-def load_adeft_models():
+def find_adeft_models():
     adeft_disambiguators = {}
     for shortform in available_adeft_models:
-        adeft_disambiguators[shortform] = load_disambiguator(shortform)
+        adeft_disambiguators[shortform] = None
     return adeft_disambiguators
 
 
+def load_adeft_models():
+    return {shortform: load_disambiguator(shortform)
+            for shortform in find_adeft_models()}
+
+
 def load_gilda_models(cutoff=0.7):
-    with gzip.open(get_gilda_models(), 'rb') as fh:
-        models_raw = pickle.load(fh)
-    models = {k: load_model_info(v['cl']) for k, v in models_raw.items()}
-    models = {k: v for k, v in models.items() if v.stats['f1']['mean'] > cutoff}
+    with gzip.open(get_gilda_models(), 'rt') as fh:
+        models = {k: load_model_info(v)
+                  for k, v in json.loads(fh.read()).items()
+                  if v['stats']['f1']['mean'] > cutoff}
     return models
